#include "filesystemmodel.h"

#include <QtCore/QUrl>
#include <QtCore/QMimeData>
#include <QApplication>
#include <io/fileimageprovider.h>
#include <io/qdriveinfo.h>

#include "filesystemmanager.h"

using namespace FileManager;

Q_GLOBAL_STATIC(FileImageProvider, imageProvider)

FileSystemModel::FileSystemModel(QObject *parent) :
    QFileSystemModel(parent),
    m_manager(FileSystemManager::instance())
{
    setObjectName("FileSystemModel");
}

FileSystemManager * FileSystemModel::fileSystemManager() const
{
    return m_manager;
}

void FileSystemModel::setFileSystemManager(FileSystemManager *manager)
{
    if (!manager)
        return;

    if (m_manager == manager)
        return;

    if (m_manager && m_manager->parent() == this)
        delete m_manager;

    m_manager = manager;
}

QVariant FileSystemModel::data(const QModelIndex &index, int role) const
{
    if (index.isValid() && index.column() == 0 && role == Qt::UserRole)
        return imageProvider()->image(QFileInfo(filePath(index)), QSize(512, 512));

    return QFileSystemModel::data(index, role);
}

bool FileSystemModel::dropMimeData(const QMimeData *data,
                  Qt::DropAction action,
                  int row,
                  int column,
                  const QModelIndex &parent
                  )
{
    Q_UNUSED(row);
    Q_UNUSED(column);
    if (!parent.isValid() || isReadOnly())
        return false;

    bool success = true;
    QString to = filePath(parent) + QDir::separator();

    QList<QUrl> urls = data->urls();

    QStringList files;
    foreach (const QUrl &url, urls) {
        QString path = url.toLocalFile();
        if (QFileInfo(path).path() != filePath(parent))
            files.append(path);
    }

    if (files.isEmpty())
        return false;

    QDriveInfo srcDrive(filePath(parent));
    QDriveInfo dstDrive(files[0]);

    //default actions
    if (srcDrive != dstDrive)
        action = Qt::CopyAction;
    else
        action = Qt::MoveAction;

    Qt::KeyboardModifiers keyboardModifiers = QApplication::keyboardModifiers();
    //change action according to the keyboard modifier, if any.
    if (keyboardModifiers.testFlag(Qt::AltModifier))
        action = Qt::CopyAction;
    else if (keyboardModifiers.testFlag(Qt::ControlModifier))
        action = Qt::MoveAction;
    else if (keyboardModifiers.testFlag(Qt::ShiftModifier))
        action = Qt::LinkAction;

    FileSystemManager * manager = fileSystemManager();

    switch (action) {
    case Qt::CopyAction: {
        manager->copy(files, to);
        break;
    }
    case Qt::LinkAction: {
        manager->link(files, to);
        break;
    }
    case Qt::MoveAction: {
        manager->move(files, to);
        break;
    }
    default:
        return false;
    }

    return success;
}
